/* C-PROGRAM FOR KERNEL REGRESSION				*/
/* prepared to be called by an Splus wrapper                    */
/* Parameters:                                                  */
/* x,y,xpred,ypred 1-D vectors of data                  	*/
/* n = number of observations					*/
/* p = number of explanatory variables  			*/
/* kernel = integer between 1 and 7 giving the kernel type      */
/*  1 = uniform                                                 */
/*  2 = cosine           					*/
/*  3 = triangular                                              */
/*  4 = Epanechnikov                 				*/
/*  5 = quartic                 				*/
/*  6 = Triweight                                               */
/*  7 = Gaussian         					*/
/*  The kernels are functions of the distance for kernel=1,2,3  */
/*  otherwise they are functions of the square distance         */
/*  They are not normalized to integrate to 1                   */
/* b = bandwidth of the kernel estimation			*/

//#include <stdlib.h>
//#include <malloc.h>
//#include <stdio.h>
//#include <math.h>

//#define PI 3.14159265


#if defined(DLL_LOAD)
#include <windows.h>
/* Standard DLL entry/exit procedure */
BOOL __stdcall
DllMain(HINSTANCE hDllInstance, DWORD dwReason, LPVOID
lpReserved)
{
        switch (dwReason) {
        case DLL_PROCESS_ATTACH:
                /* initialization code here */
                break;
        case DLL_PROCESS_DETACH:
                /* clean-up code here */
                break;
        }
        return(TRUE);
}
#endif

#include <S.h>
#include <newredef.h>

#define error(msg)	{fprintf(stderr, (msg)); return;}

double kernel1 (double u)
{
  double w;
  if ( u < 1)  w=.5; else w=0.0;
  return(w);
}

double kernel2 (double u)
{
  double w;
  if ( u < 1) w= cos(PI/2.0 * u); else w=0.0;
  return(w);
}

double kernel3 (double u)
{
  double w;
  if ( u < 1) w= 1-u; else w=0.0;
  return(w);
}

double kernel4 (double u)
{
  double w;
  if ( u < 1) w= 1-u; else w=0.0;
  return(w);
}

double kernel5 (double u)
{
  double w;
  if ( u < 1) w= (1-u)*(1-u); else w=0.0;
  return(w);
}

double kernel6 (double u)
{
  double w;
  if ( u < 1) w= (1-u)*(1-u)*(1-u); else w=0.0;
  return(w);
}

double kernel7 (double u)
{
  return(exp(-0.5 * u));
}

void kreg(double *x, double *y, double *xpred, double *ypred, int *pn, 
int *pp, int *pnpred, int *pkernel, double *pb)
{
  double **X, **XPRED;
  double num, den;
  double distance,weight;
  int n,p,npred,kernel;
  int i,j,k;
  double (*kernelproc)();
  double b;

  n = *pn;
  npred = *pnpred;
  p = *pp;
  kernel = *pkernel;
  b = *pb;
  if(kernel <= 3) b = sqrt(b);
/*
printf("bandwidth = %f\n",b);
*/

  if(!(X = (double **)malloc(sizeof(double *) * n)))
    error("Memory allocation failed for X in kreg.c \n");
  if(!(XPRED = (double **)malloc(sizeof(double *) * npred)))
    error("Memory allocation failed for XPRED in kreg.c \n");

  for(i = 0; i < n; i++) 
    if(!(X[i] = (double *)malloc(sizeof(double) * p)))
      error("Memory allocation failed for X[] in kreg.c \n");

  for(i = 0; i < npred; i++) 
    if(!(XPRED[i] = (double *)malloc(sizeof(double) * p)))
      error("Memory allocation failed for XPRED[] in kreg.c \n");

  for(j = 0; j < p; j++)   
    for( i = 0; i < n; i++)
      X[i][j] = (double)(x[j * n + i]);
  for(j = 0; j < p; j++)   
    for( i = 0; i < npred; i++)
      XPRED[i][j] = (double)(xpred[j * npred + i]);

  switch (kernel)
  {
    case 1 :  kernelproc = kernel1; break;
    case 2 :  kernelproc = kernel2; break;
    case 3 :  kernelproc = kernel3; break;
    case 4 :  kernelproc = kernel4; break;
    case 5 :  kernelproc = kernel5; break;
    case 6 :  kernelproc = kernel6; break;
    case 7 :  kernelproc = kernel7; break;
    default:  printf(" Error in the choice of kernel"); return;
  }

  for (k=0; k<npred; k++)
  {

/*
printf("k = %d\n",k);
*/
       num = 0.0;
       den = 0.0;
       for (i=0; i<n; i++)
       {
          distance = 0.0;
          for (j=0; j<p; j++)
              distance += (X[i][j]-XPRED[k][j])*(X[i][j]-XPRED[k][j]);
	  if(kernel <= 3)
	        distance = sqrt(distance);
/*
printf("distance = %f\n",distance);
*/
          weight = kernelproc(distance/b);
/*
printf("weight = %f\n",weight);
*/
          num += weight * y[i];
/*
printf("numerator = %f\n",num);
*/
          den += weight;
/*
printf("denominator = %f\n",den);
*/
       }
/*
printf("final numerator = %f\n",num);
printf("final denominator = %f\n",den);
*/        
    if(den == 0)
          error("Division by zero in the kernel computation. \n");
    ypred[k] = num/den;

/*
printf("ypred[k] = %f\n",ypred[k]);
*/


  }

  for(i = 0; i < npred; i++)
      free(XPRED[i]);
  for(i = 0; i < n; i++) 
      free(X[i]);

  free(XPRED);
  free(X);
}

/* end */

  
